#include <gmp.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#define __STDC_FORMAT_MACROS
#include <inttypes.h>

// Define these function's to make it more readable
#define mpz_rshift(x, y) mpz_tdiv_q_2exp(x, y, 1)
#define mpz_lshift(x, y) mpz_mul_2exp(x, y, 1)

#define mpz_dec(x) mpz_sub_ui(x, x, 1);
#define mpz_inc(x) mpz_add_ui(x, x, 1);

int main (int argc, char *argv[])
{
	gmp_randstate_t state;
	mpz_t n, e, p, q, pq, d, tap, m, temp;
	unsigned int i;
	uint64_t total = 0;
	int resuming = 0;
	time_t curtime;

	// Initialze and set default values for 'e' and 'n'
	mpz_init_set_si(e, 65537);
	//mpz_init_set_str(n, "c3b3a7015c04299ff3a25f104e2285c1ec2d55471e6208959d0f6981b2fa2c6d3e316f9364d5eb5c7789e142b75bfaf402e7e02fac0cb09f6419db1f44679f8bbcca142f1d312feb095708ef175a4ef80271321e7240f0d854c90a74fc59209cdf80aa8f85ae3b948a3ce55c69cd050098d5a79aebbc241cc642b106b1af2cb7", 16);
	mpz_init_set_str(n, "aba7f0b8c7feb6e33438af5c25c67389eaf4d73f80cd0a37922493431cde03b34da448bdb05387cf7a8c59ee12d9613429a2b07ea385752f079892da1ae76c2b158f2d7169aae066432fe44f797df39dd6a0d7b2e2091281b30efac247c51576ebc93ec456de2e27d36b713844336b65af67ee58e6107a6a1deb954a91095295", 16);
	//mpz_init_set_str(n, "b15e01c47c421be62f4e769b3ac98f4f983a820b0c181e35715d84a4f1acf0527eeddfabf9f66e73bedb55376e22f860c34dc70ce239157297056d4ecf46535778c3917647b5a6bb9c5638cdeff3e309ff66878fd4f233cf157d7af4136f307df90ec6ae6eaff6bead6d52f423a37dac59ff38ae876008103728f2bd674e858f", 16);
	mpz_init(p);

	// Loop through arguments
	if (argc > 0)
	{
		char *ptr;

		while (--argc)
		{
			argv++;

			// If the user has defined a value for 'e', use it rather than default
			if(!strcmp(*argv, "-e"))
			{
				argc--;
				argv++;
				ptr = *argv;
				if (ptr[0] == '0' && ptr[1] == 'x')
				{
					// Assume hex number.
					ptr += 2;
				}
				mpz_set_str(e, ptr, 16);
			}

			// If the user has requested help on the command, give it to them
			else if(!strcmp(*argv, "-h") || !strcmp(*argv, "--help"))
			{
				printf("Usage: client [-h] [-e e] [-r p] [n]\n");
				return 0;
			}

			// If the user has defined a resume value for 'p', use it.
			else if(!strcmp(*argv, "-r"))
			{
				resuming = 1;
				argc--;
				argv++;
				ptr = *argv;
				if (ptr[0] == '0' && ptr[1] == 'x')
				{
					// Assume hex number.
					ptr += 2;
				}
				mpz_set_str(p, ptr, 16);
			}

			// If none of the above conditions are satisfied, assume the user entered a value for n
			else
			{
				argc--;
				argv++;
				ptr = *argv;
				if (ptr[0] == '0' && ptr[1] == 'x')
				{
					// Assume hex number.
					ptr += 2;
				}
				mpz_set_str(n, ptr, 16);
			}
		}
	}

	// Some error checking
	if (!mpz_cmp_si(e, -1) || !mpz_cmp_si(e, 0) || !mpz_cmp_si(n, -1) || !mpz_cmp_si(n, 0)|| !mpz_cmp_si(p, -1))
	{
		printf("::: ERROR: Invalid number\n");
		printf(":: Please enter valid integers in base 16 representation, with or without a leading '0x'\n");
		return 1;
	}

	// Initialize the other variables needed
	// 512-bit Galois LFSR generator: bits 511, 509, 506 and 503 set.
	mpz_init_set_str(tap, "a4800000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000", 16);
	mpz_init_set_si(m, 1);
	mpz_init(temp);

redo:
	// Initialize p with a random 512-bit value if we're not resuming.
	curtime = time(NULL);
	gmp_randinit_mt(state);
	gmp_randseed_ui(state, (unsigned long)curtime);
	if (!resuming)
	{
		mpz_urandomb(p, state, 512);
		i = 0;
	}
	else
	{
		i = -1;
	}
	gmp_randclear(state);
	//mpz_init_set_str(p, "4c2e0873", 16);
	//mpz_init_set_str(p, "a8ba4642eeece501884810138ce6228692d08e98acda06ba6386a2cb13da36aafe0617c25f19dc3b5f485c8f9e62dfdde298cd5be67573fa24fb1f96617a6e1f", 16);
	//mpz_init_set_str(p, "9e372f4ab513eb13a021bb1334236af850a17af4c96fe8c613cf4f4fbb8b63b81ef78d240df89593b0ac49b76d6bad91711c4eeb1f2a8ff06909ecef68d9735", 16);

	printf("Program started at %ld\n", (unsigned long)curtime);
	printf("Factoring n = ");
	mpz_out_str(stdout, 16, n);
	printf("\n");
	gmp_printf("p=%#Zx\n", p);

	for (;;)
	{
		// If 'p' is prime, right shift by one, otherwise do the same thing and then xor the number
		if (!mpz_tstbit(p, 0))
		{
			mpz_rshift(temp, p);
			mpz_xor(p, temp, tap);
			// Slower.
			/*mpz_combit (temp, 511);
			mpz_combit (temp, 509);
			mpz_combit (temp, 506);
			mpz_combit (temp, 503);
			mpz_swap(temp, p);*/
		}
		else
		{
			total++;

			if (i == 65535)
			{
				fprintf(stdout, "%" PRIu64 " odd integers\n", total);
				gmp_printf("p=%#Zx\n", p);
				i = -1;
			}

			// NOTE ON PERFORMANCE (Core 2 Duo T7200 @ 2 GHz and T7600 @ 2.33 GHz): the program runs ~80 times slower (!!) if
			// the pseudo-primality testing (without any Miller-Rabin tests, which slow down a bit more) is activated.
			// 
			// That pretty much says it all: even if the density of primes on 512-bit integers is, more or less,
			// 1 out of 200 integers, it makes no sense to run any primality test.
			// The only thing we can do to speed up a bit, is trial division with the smallest prime divisors.
			//if (mpz_probab_prime_p(p, 0))
			if (   !mpz_divisible_ui_p(p, 3)
			    && !mpz_divisible_ui_p(p, 5)
			    && !mpz_divisible_ui_p(p, 7)
			    && !mpz_divisible_ui_p(p, 11)
			    && !mpz_divisible_ui_p(p, 13)
			   )
			{

				if (mpz_divisible_p(n, p))
				{
					break;
				}
			}
			mpz_rshift(temp, p);
			mpz_swap(temp, p);
			i++;
		}
	}

	if (!mpz_cmp_si(p, 1))
	{
		printf("::: p = 1, restarting main loop\n");
		resuming = 0;
		goto redo;
	}

	// If we make it this far, 'p' and 'q' were found
	// So let's print the good news!
	printf("\n::: Found 'p' and 'q'!!!\n");
	printf(":: Computing private key...\n");

	// Calculate the private key.
	mpz_init(q);
	mpz_init(pq);
	mpz_init(d);
	mpz_tdiv_q(q, n, p);	// Divide 'n' by 'p' and store the result in 'q'. (The division truncates)
	mpz_sub_ui(pq, p, 1);	// Subtruct one from 'p' and store the result in 'pq'
	mpz_dec(q);				// Decrement 'q' by one
	mpz_mul(pq, pq, q);		// Get the result of '(q-1)(p-1)'
	mpz_inc(q);				// Increment 'q'
	mpz_invert(d, e, pq);	// Get the mod inverse of 'e' and 'pq' and store it in 'd'. (This get's the private key)

	// Finally, output all the values
	gmp_printf(": n = %#Zx\n", n);
	gmp_printf(": p = %#Zx\n", p);
	gmp_printf(": q = %#Zx\n", q);
	gmp_printf(": e = %#Zx\n", e);
	gmp_printf(": d = %#Zx\n", d);

	printf(":: Saving to file 'key.txt'...\n");

	FILE *fp;
	fp = fopen("key.txt", "w");

	// Check that we were able to open a file stream
	if(fp != NULL)
	{
		// If we were, start saving the data
		fprintf(fp, "n = ");
		mpz_out_str(fp, 16, n);

		fprintf(fp, "\np = ");
		mpz_out_str(fp, 16, p);

		fprintf(fp, "\nq = ");
		mpz_out_str(fp, 16, q);

		fprintf(fp, "\ne = ");
		mpz_out_str(fp, 16, e);

		fprintf(fp, "\nd = ");
		mpz_out_str(fp, 16, d);

		fclose(fp);
	}
	else
	{
		// Otherwise, alert the user that the program will not quit and
		// will stay open until they close it
		printf("::: ERROR: Unable to save to file!!!\n");
		printf(":: This program will stay open until explicitly closed.");

		// Yes, I know I could have done while(1); but I just felt like putting this in here
		__asm("loop:\nnop\njmp loop");
	}

	return 0;
}
